use std::{
    string::String,
    sync::{Arc, Mutex},
};

use dashmap::{DashMap, DashSet};
use rowan::ast::{AstNode, AstPtr};

use crate::SyntaxNode;

use super::{
    files::{FileId, Loc},
    nodes,
    registry::ItemPath,
    world_creation_pool::{wait_blocker::WaitBlocker, Task, WorkerCtx},
};

pub struct Module {
    /// is none if this is the root module of the entry point file
    decl: Option<Loc<nodes::Mod>>,
    body: ModuleBody,
    own_path: ItemPath,
    child_modules: Arc<DashMap<String, Arc<Module>>>,
    child_defs: Arc<DashMap<String, Def>>,
}

impl Module {
    pub fn parse_mod_body(
        ctx: &WorkerCtx,
        tree: SyntaxNode,
        path: ItemPath,
        file: FileId,
    ) -> (DashMap<String, Arc<Mutex<Option<Arc<Module>>>>>, Vec<Task>) {
        let children: Vec<(String, Arc<Mutex<Option<Arc<Module>>>>, nodes::Mod)> = tree
            .children()
            .filter_map(|c| nodes::Mod::cast(c))
            .map(|m| {
                let name = nodes::ModName::cast(m.syntax().first_child().unwrap())
                    .expect("ModName should always be first child of Mod");
                (
                    name.syntax().text().to_string(),
                    Arc::new(Mutex::new(None)),
                    m,
                )
            })
            .collect();
        let retv: DashMap<String, Arc<Mutex<Option<Arc<Module>>>>> = children
            .iter()
            .map(|(name, mod_r, _)| (name.to_owned(), mod_r.clone()))
            .collect();
        let tasks = children
            .into_iter()
            .map(|(name, mod_r, loc)| {
                let mut path = path.clone();
                path.push(name);
                Task::ParseMod(Loc::new(loc, file), mod_r, path)
            })
            .collect::<Vec<_>>();

        (retv, tasks)
    }
    pub fn parse_file0(
        ctx: &mut WorkerCtx,
        file: FileId,
        path: ItemPath,
        decl: Option<Loc<nodes::Mod>>,
    ) {
        let tree = ctx.get_tree(file);

        let (retv, tasks) = Self::parse_mod_body(ctx, tree, path.clone(), file);

        ctx.send_tasks(
            Task::ParseFileMod1 {
                file,
                decl,
                ret: retv.into(),
                path,
            },
            tasks,
        )
    }

    pub fn parse_mod(
        ctx: &mut WorkerCtx,
        loc: Loc<nodes::Mod>,
        path: ItemPath,
        ret: Arc<Mutex<Option<Module>>>,
        blocker: WaitBlocker,
    ) {
        let mod_decl = ctx.resolve_loc(loc);
        let children = mod_decl.syntax().children().collect::<Vec<_>>();

        if children.len() == 1 {
            // TODO: file mod
            todo!()
        } else if children.len() == 2 {
            // inline mod
        }
    }

    pub fn parse_inline_mod(
        ctx: &mut WorkerCtx,
        decl: Loc<nodes::Mod>,
        path: ItemPath,
        ret: Arc<Mutex<Option<Arc<Module>>>>,
        blocker: WaitBlocker,
    ) {
        let mod_decl_children = ctx
            .resolve_loc(decl.clone())
            .syntax()
            .children()
            .collect::<Vec<_>>();

        assert_eq!(mod_decl_children.len(), 2);

        let name = nodes::ModName::cast(mod_decl_children[0].clone()).unwrap();
        let body = nodes::ModBody::cast(mod_decl_children[1].clone()).unwrap();

        let (retv, tasks) =
            Self::parse_mod_body(ctx, body.syntax().clone(), path.clone(), decl.file());

        ctx.send_tasks(
            Task::CompleteMod {
                ret_self: ret,
                body: ModuleBody::Inline(Loc::new(body, decl.file())),
                decl: Some(decl),
                path,
                child_mods: retv.into(),
                blocker,
            },
            tasks,
        )
    }

    pub fn complete_mod(
        ctx: &mut WorkerCtx,
        decl: Option<Loc<nodes::Mod>>,
        ret_self: Arc<Mutex<Option<Arc<Module>>>>,
        body: ModuleBody,
        path: ItemPath,
        child_mods: Arc<DashMap<String, Arc<Mutex<Option<Arc<Module>>>>>>,
        blocker: WaitBlocker,
    ) {
        assert!(child_mods
            .iter()
            .all(|item| item.value().lock().unwrap().is_some()));

        let module = Arc::new(Module {
            decl,
            body,
            own_path: path,
            child_modules: Arc::new(
                Arc::<_>::unwrap_or_clone(child_mods)
                    .into_iter()
                    .map(|(name, module)| {
                        let module = module.lock().unwrap().take().unwrap();
                        (name, module)
                    })
                    .collect(),
            ),
            child_defs: Arc::default(),
        });

        let mut ret_self = ret_self.lock().unwrap();
        *ret_self = Some(module);
        drop(blocker);
    }
}

struct Def;

pub enum ModuleBody {
    Inline(Loc<nodes::ModBody>),
    File(FileId),
}
