use std::{
    collections::HashMap,
    path::PathBuf,
    sync::{Arc, Mutex},
    thread,
};

use crossbeam::channel::{Receiver, Sender};
use dashmap::DashMap;
use rowan::ast::{AstNode, AstPtr};

use crate::{Lang, SyntaxNode};

use self::wait_blocker::WaitBlocker;

use super::{
    error::Error,
    files::{FileId, Files, Loc},
    modules::{Module, ModuleBody},
    nodes,
    registry::{ItemPath, Registry},
};

pub mod wait_blocker;

pub(super) struct WorldCreationPool {
    workers: Vec<Worker>,
    tx: Sender<Job>,
}

impl WorldCreationPool {
    pub fn new(files: Files, reg: Registry) -> Self {
        let mut workers = Vec::new();
        let (tx, rx) = crossbeam::channel::unbounded();

        for i in 0..num_cpus::get() {
            workers.push(Worker::new(
                i,
                files.clone(),
                reg.clone(),
                tx.clone(),
                rx.clone(),
            ))
        }

        Self { workers, tx }
    }
}

enum Job {
    WaitingFor(WaitBlocker, Task),
    Awaited(WaitBlocker, Task),
}

pub enum Task {
    ParseMod(Loc<nodes::Mod>, Arc<Mutex<Option<Arc<Module>>>>, ItemPath),
    ParseFileMod0 {
        file: FileId,
        decl: Option<Loc<nodes::Mod>>,
        path: ItemPath,
    },
    ParseFileMod1 {
        file: FileId,
        decl: Option<Loc<nodes::Mod>>,
        path: ItemPath,
        ret: Arc<DashMap<String, Arc<Mutex<Option<Arc<Module>>>>>>,
    },
    CompleteMod {
        ret_self: Arc<Mutex<Option<Arc<Module>>>>,
        decl: Option<Loc<nodes::Mod>>,
        body: ModuleBody,
        path: ItemPath,
        child_mods: Arc<DashMap<String, Arc<Mutex<Option<Arc<Module>>>>>>,
        blocker: WaitBlocker,
    },
    OpenFile(PathBuf),
}

struct Worker {
    id: usize,
    // thead: thread::JoinHandle<Vec<Error>>,
    thread: thread::JoinHandle<()>,
}

pub struct WorkerCtx {
    errors: Vec<Error>,
    files: Files,
    local_files: HashMap<PathBuf, SyntaxNode>,
    pub reg: Registry,
    tx: Sender<Job>,
}

impl WorkerCtx {
    pub fn get_tree(&mut self, id: FileId) -> SyntaxNode {
        let p = self.files.get_path(id);
        if self.local_files.contains_key(&p) {
            self.local_files[&p].clone()
        } else {
            let f = self.files.get(id);
            let lst = SyntaxNode::new_root(f.lst.read().unwrap().clone());
            self.local_files.insert(p, lst.clone());
            lst
        }
    }

    pub fn resolve_loc<N: AstNode<Language = Lang>>(&mut self, loc: Loc<N>) -> N {
        let f = self.get_tree(loc.file());

        loc.syntax().to_node(&f)
    }

    pub fn send_tasks(&self, task: Task, dependencies: Vec<Task>) {
        let blocker = WaitBlocker::new();
        for dep_task in dependencies {
            self.tx
                .send(Job::Awaited(blocker.clone(), dep_task))
                .unwrap();
        }
        self.tx.send(Job::WaitingFor(blocker, task)).unwrap();
    }
}

impl Worker {
    fn new(id: usize, files: Files, reg: Registry, sender: Sender<Job>, rx: Receiver<Job>) -> Self {
        let thread_handle = thread::spawn(move || {
            let ctx = WorkerCtx {
                errors: Vec::new(),
                local_files: HashMap::new(),
                files,
                reg,
                tx: sender,
            };

            for job in &rx {
                // if matches!(job, Job::WithCond(_, _)) {

                // }
                match job {
                    Job::WaitingFor(blocker, task) => {
                        if blocker.is_ready() {
                            Self::do_task(&ctx, task, None)
                        } else if rx.is_empty() {
                            if let Some(blocker) =
                                blocker.wait_for(std::time::Duration::from_millis(50))
                            {
                                ctx.tx.send(Job::WaitingFor(blocker, task)).unwrap();
                            } else {
                                Self::do_task(&ctx, task, None)
                            }
                        } else {
                            ctx.tx.send(Job::WaitingFor(blocker, task)).unwrap();
                        }
                    }
                    Job::Awaited(blocker, task) => {
                        Self::do_task(&ctx, task, Some(blocker.clone()));
                        drop(blocker)
                    }
                }

                // if let Job::WithCond(blocker, job_inner) = job_msg {
                //     if blocker.is_ready() {
                //         job = Box::<Job>::into_inner(job_inner);
                //     } else if rx.is_empty() {
                //         if let Some(blocker) =
                //             blocker.wait_for(std::time::Duration::from_millis(50))
                //         {
                //             job = Job::WithCond(blocker, job_inner);
                //         } else {
                //             job = Box::<Job>::into_inner(job_inner);
                //         }
                //     } else {
                //         job = Job::WithCond(blocker, job_inner);
                //     }
                // } else {
                //     job = job_msg;
                // }

                // match job {
                //     Job::ParseFileMod { file, decl } => todo!(),
                //     Job::OpenFile(path) => todo!(),
                //     Job::WithCond(blocker, job) => {
                //         ctx.tx.send(Job::WithCond(blocker, job)).unwrap()
                //     }
                // }
            }
        });

        Self {
            id,
            thread: thread_handle,
        }
    }

    fn do_task(ctx: &WorkerCtx, task: Task, blocker: Option<WaitBlocker>) {
        match task {
            Task::ParseMod(_, _, _) => todo!(),
            Task::ParseFileMod0 { file, decl, path } => todo!(),
            Task::ParseFileMod1 {
                file,
                decl,
                path,
                ret,
            } => todo!(),
            Task::CompleteMod {
                ret_self,
                decl,
                body,
                path,
                child_mods,
                blocker,
            } => todo!(),
            Task::OpenFile(_) => todo!(),
        }
    }
}
