use std::{
    sync::{Arc, Condvar, Mutex},
    time::Duration,
};

/// like a WaitGroup from crossbeam, but can also just check if it's the last one
pub struct WaitBlocker {
    inner: Arc<Inner>,
}

struct Inner {
    waiting: Mutex<usize>,
    cvar: Condvar,
}

impl WaitBlocker {
    pub fn new() -> Self {
        Self {
            inner: Arc::new(Inner {
                waiting: Mutex::new(1),
                cvar: Condvar::new(),
            }),
        }
    }

    pub fn wait(self) {
        if *self.inner.waiting.lock().unwrap() == 1 {
            return;
        }

        let inner = self.inner.clone();
        drop(self);

        inner
            .cvar
            .wait_while(inner.waiting.lock().unwrap(), |w| *w > 0);
    }

    pub fn wait_for(self, dur: Duration) -> Option<Self> {
        if *self.inner.waiting.lock().unwrap() == 1 {
            return None;
        }

        let inner = self.inner.clone();
        drop(self);

        let (_, timeout_res) = inner
            .cvar
            .wait_timeout_while(inner.waiting.lock().unwrap(), dur, |w| *w > 0)
            .unwrap();

        if timeout_res.timed_out() {
            None
        } else {
            {
                let mut w = inner.waiting.lock().unwrap();
                *w += 1;
            }
            Some(Self { inner })
        }
    }

    pub fn is_ready(&self) -> bool {
        *self.inner.waiting.lock().unwrap() == 1
    }
}

impl Clone for WaitBlocker {
    fn clone(&self) -> Self {
        let mut w = self.inner.waiting.lock().unwrap();
        *w += 1;
        drop(w);
        Self {
            inner: self.inner.clone(),
        }
    }
}

impl Drop for WaitBlocker {
    fn drop(&mut self) {
        let mut w = self.inner.waiting.lock().unwrap();
        *w -= 1;
        if *w == 0 {
            self.inner.cvar.notify_all()
        }
    }
}
